/**
  *  \file t_format.cc
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/format.h"
#include "cpluslib/string.h"

#include "tests.h"

void
TestFormat::testConv()
{
    // This mainly tests correct initialisation/conversion
    {
        string_t s = format("foo%d") << 42;
        TS_ASSERT_EQUALS(s, "foo42");
    }
    {
        string_t s = format("foo%d", 42);
        TS_ASSERT_EQUALS(s, "foo42");
    }
}

void
TestFormat::testInt()
{
    TS_ASSERT_EQUALS(string_t(format("dezimal            = `%d'") << 42),        "dezimal            = `42'");
    TS_ASSERT_EQUALS(string_t(format("hex                = `%x'") << 42),        "hex                = `2a'");
    TS_ASSERT_EQUALS(string_t(format("oct                = `%o'") << 42),        "oct                = `52'");
    TS_ASSERT_EQUALS(string_t(format("alt.hex            = `%#x'") << 42),       "alt.hex            = `0x2a'");
    TS_ASSERT_EQUALS(string_t(format("alt.oct            = `%#o'") << 42),       "alt.oct            = `052'");
    TS_ASSERT_EQUALS(string_t(format("alt.dezimal        = `%#d'") << 42),       "alt.dezimal        = `42'");
    TS_ASSERT_EQUALS(string_t(format("group dez          = `%'d'") << 42),       "group dez          = `42'");
    TS_ASSERT_EQUALS(string_t(format("group oct          = `%'o'") << 42),       "group oct          = `52'");
    TS_ASSERT_EQUALS(string_t(format("10-dez             = `%10d'") << 42),      "10-dez             = `        42'");
    TS_ASSERT_EQUALS(string_t(format("left 10-dez        = `%-10d'") << 42),     "left 10-dez        = `42        '");
    TS_ASSERT_EQUALS(string_t(format("left group 10-dez  = `%'-10d'") << 42),    "left group 10-dez  = `42        '");
    TS_ASSERT_EQUALS(string_t(format("group 10-dez       = `%'10d'") << 42),     "group 10-dez       = `        42'");
    TS_ASSERT_EQUALS(string_t(format("zpad 10-dez        = `%010d'") << 42),     "zpad 10-dez        = `0000000042'");
    TS_ASSERT_EQUALS(string_t(format("sign zpad 10-dez   = `%+010d'") << 42),    "sign zpad 10-dez   = `+000000042'");
    TS_ASSERT_EQUALS(string_t(format("group 10-dez       = `%'10d'") << 1024),   "group 10-dez       = `     1,024'");
    TS_ASSERT_EQUALS(string_t(format("group 10-dez       = `%'10d'") << 45054),  "group 10-dez       = `    45,054'");
    TS_ASSERT_EQUALS(string_t(format("group 10-hex       = `%'10X'") << 45054),  "group 10-hex       = `     A,FFE'");
    TS_ASSERT_EQUALS(string_t(format("group alt 10-hex   = `%#'10X'") << 45054), "group alt 10-hex   = `   0XA,FFE'");
    TS_ASSERT_EQUALS(string_t(format("group alt 10 dez   = `%#'10X'") << 999),   "group alt 10 dez   = `     0X3E7'");
    TS_ASSERT_EQUALS(string_t(format("group alt neg dez  = `%#'d'") << -999),    "group alt neg dez  = `-999'");
    TS_ASSERT_EQUALS(string_t(format("group alt neg10dez = `%#'10d'") << -999),  "group alt neg10dez = `      -999'");
    TS_ASSERT_EQUALS(string_t(format("zpad alt neg10dez  = `%#010d'") << -999),  "zpad alt neg10dez  = `-000000999'");
    TS_ASSERT_EQUALS(string_t(format("neg 10 dez         = `%#10d'") << -4711),  "neg 10 dez         = `     -4711'");
    TS_ASSERT_EQUALS(string_t(format("left neg 10 dez    = `%-#10d'") << -4711), "left neg 10 dez    = `-4711     '");
    TS_ASSERT_EQUALS(string_t(format("grp neg dez        = `%'d'") << -4711),    "grp neg dez        = `-4,711'");

    static const char* str = "cond = `%d %1{item for %d dollar%!1{s%}%|items for %d dollar%!1{s%} each%}'";

    TS_ASSERT_EQUALS(string_t(format(str) << 0 << 0), "cond = `0 items for 0 dollars each'");
    TS_ASSERT_EQUALS(string_t(format(str) << 0 << 1), "cond = `0 items for 1 dollar each'");
    TS_ASSERT_EQUALS(string_t(format(str) << 0 << 2), "cond = `0 items for 2 dollars each'");
    TS_ASSERT_EQUALS(string_t(format(str) << 1 << 0), "cond = `1 item for 0 dollars'");
    TS_ASSERT_EQUALS(string_t(format(str) << 1 << 1), "cond = `1 item for 1 dollar'");
    TS_ASSERT_EQUALS(string_t(format(str) << 1 << 2), "cond = `1 item for 2 dollars'");
    TS_ASSERT_EQUALS(string_t(format(str) << 2 << 0), "cond = `2 items for 0 dollars each'");
    TS_ASSERT_EQUALS(string_t(format(str) << 2 << 1), "cond = `2 items for 1 dollar each'");
    TS_ASSERT_EQUALS(string_t(format(str) << 2 << 2), "cond = `2 items for 2 dollars each'");
}

void
TestFormat::testFP()
{
    /* floating-point tests */
    TS_ASSERT_EQUALS(string_t(format("format e           = `%e'") << 4711.0),     "format e           = `4.711000e+03'");
    TS_ASSERT_EQUALS(string_t(format("format f           = `%f'") << 4711.0),     "format f           = `4711.000000'");
    TS_ASSERT_EQUALS(string_t(format("format g           = `%g'") << 4711.0),     "format g           = `4711'");
    TS_ASSERT_EQUALS(string_t(format("format e one       = `%e'") << 1.0),        "format e one       = `1.000000e+00'");
    TS_ASSERT_EQUALS(string_t(format("format f one       = `%f'") << 1.0),        "format f one       = `1.000000'");
    TS_ASSERT_EQUALS(string_t(format("format g one       = `%g'") << 1.0),        "format g one       = `1'");
    TS_ASSERT_EQUALS(string_t(format("format e zero      = `%e'") << 0.0),        "format e zero      = `0.000000e+00'");
    TS_ASSERT_EQUALS(string_t(format("format f zero      = `%f'") << 0.0),        "format f zero      = `0.000000'");
    TS_ASSERT_EQUALS(string_t(format("format g zero      = `%g'") << 0.0),        "format g zero      = `0'");
    TS_ASSERT_EQUALS(string_t(format("format e tiny      = `%e'") << 0.00000123), "format e tiny      = `1.230000e-06'");
    TS_ASSERT_EQUALS(string_t(format("format f tiny      = `%f'") << 0.00000123), "format f tiny      = `0.000001'");
    TS_ASSERT_EQUALS(string_t(format("format g tiny      = `%g'") << 0.00000123), "format g tiny      = `1.23e-06'");
    TS_ASSERT_EQUALS(string_t(format("format .2g         = `%.2g'") << 12345.0),  "format .2g         = `1.23e+04'");

    /* more floating-point tests */
    TS_ASSERT_EQUALS(string_t(format("format .2f .999     = `%.2f'") << 0.999),   "format .2f .999     = `1.00'");
    TS_ASSERT_EQUALS(string_t(format("format .2e .999     = `%.2e'") << 0.999),   "format .2e .999     = `9.99e-01'");
    TS_ASSERT_EQUALS(string_t(format("format .2f 99.999   = `%.2f'") << 99.999),  "format .2f 99.999   = `100.00'");
    TS_ASSERT_EQUALS(string_t(format("format .2e 99.999   = `%.2e'") << 99.999),  "format .2e 99.999   = `1.00e+02'");
    TS_ASSERT_EQUALS(string_t(format("format .5f 99.999   = `%.5f'") << 99.999),  "format .5f 99.999   = `99.99900'");
    TS_ASSERT_EQUALS(string_t(format("format .5e 99.999   = `%.5e'") << 99.999),  "format .5e 99.999   = `9.99990e+01'");
    TS_ASSERT_EQUALS(string_t(format("format .2f 25       = `%.2f'") << 25.0),    "format .2f 25       = `25.00'");
    TS_ASSERT_EQUALS(string_t(format("format .2e 25       = `%.2e'") << 25.0),    "format .2e 25       = `2.50e+01'");
    TS_ASSERT_EQUALS(string_t(format("format .2f .25      = `%.2f'") << 0.25),    "format .2f .25      = `0.25'");
    TS_ASSERT_EQUALS(string_t(format("format .2e .25      = `%.2e'") << 0.25),    "format .2e .25      = `2.50e-01'");

    TS_ASSERT_EQUALS(string_t(format("format g hundred    = `%g'") << 100.0),     "format g hundred    = `100'");
    TS_ASSERT_EQUALS(string_t(format("format f hundred    = `%.0f'") << 100.0),   "format f hundred    = `100'");
    TS_ASSERT_EQUALS(string_t(format("format f hundred    = `%.1f'") << 100.0),   "format f hundred    = `100.0'");

    // Rounding
    TS_ASSERT_EQUALS(string_t(format("%.1f", 0.003003003)), "0.0");
    TS_ASSERT_EQUALS(string_t(format("%.1f", 0.99)), "1.0");
    TS_ASSERT_EQUALS(string_t(format("%.1f", 99.99)), "100.0");
}

/** Test that strings formatted as integers set flags. */
void
TestFormat::testIntStr()
{
    static const char* str = "cond = `%d %1{item for %d dollar%!1{s%}%|items for %d dollar%!1{s%} each%}'";

    // Using const char*:
    TS_ASSERT_EQUALS(string_t(format(str) << "0" << "0"), "cond = `0 items for 0 dollars each'");
    TS_ASSERT_EQUALS(string_t(format(str) << "0" << "1"), "cond = `0 items for 1 dollar each'");
    TS_ASSERT_EQUALS(string_t(format(str) << "0" << "2"), "cond = `0 items for 2 dollars each'");
    TS_ASSERT_EQUALS(string_t(format(str) << "1" << "0"), "cond = `1 item for 0 dollars'");
    TS_ASSERT_EQUALS(string_t(format(str) << "1" << "1"), "cond = `1 item for 1 dollar'");
    TS_ASSERT_EQUALS(string_t(format(str) << "1" << "2"), "cond = `1 item for 2 dollars'");
    TS_ASSERT_EQUALS(string_t(format(str) << "2" << "0"), "cond = `2 items for 0 dollars each'");
    TS_ASSERT_EQUALS(string_t(format(str) << "2" << "1"), "cond = `2 items for 1 dollar each'");
    TS_ASSERT_EQUALS(string_t(format(str) << "2" << "2"), "cond = `2 items for 2 dollars each'");

    TS_ASSERT_EQUALS(string_t(format(str) << "0000000000000" << "0000000000000"), "cond = `0000000000000 items for 0000000000000 dollars each'");
    TS_ASSERT_EQUALS(string_t(format(str) << "0000000000000" << "0000000000001"), "cond = `0000000000000 items for 0000000000001 dollar each'");
    TS_ASSERT_EQUALS(string_t(format(str) << "0000000000000" << "0000000000002"), "cond = `0000000000000 items for 0000000000002 dollars each'");
    TS_ASSERT_EQUALS(string_t(format(str) << "000000000001x" << "0004294967297"), "cond = `000000000001x item for 0004294967297 dollars'");

    TS_ASSERT_EQUALS(string_t(format(str) << "000,000,000,1" << ",,,,,,0,,0,,,"), "cond = `000,000,000,1 item for ,,,,,,0,,0,,, dollars'");
    
    // Using string_t:
    TS_ASSERT_EQUALS(string_t(format(str) << string_t("0") << string_t("0")), "cond = `0 items for 0 dollars each'");
    TS_ASSERT_EQUALS(string_t(format(str) << string_t("0") << string_t("1")), "cond = `0 items for 1 dollar each'");
    TS_ASSERT_EQUALS(string_t(format(str) << string_t("0") << string_t("2")), "cond = `0 items for 2 dollars each'");
    TS_ASSERT_EQUALS(string_t(format(str) << string_t("1") << string_t("0")), "cond = `1 item for 0 dollars'");
    TS_ASSERT_EQUALS(string_t(format(str) << string_t("1") << string_t("1")), "cond = `1 item for 1 dollar'");
    TS_ASSERT_EQUALS(string_t(format(str) << string_t("1") << string_t("2")), "cond = `1 item for 2 dollars'");
    TS_ASSERT_EQUALS(string_t(format(str) << string_t("2") << string_t("0")), "cond = `2 items for 0 dollars each'");
    TS_ASSERT_EQUALS(string_t(format(str) << string_t("2") << string_t("1")), "cond = `2 items for 1 dollar each'");
    TS_ASSERT_EQUALS(string_t(format(str) << string_t("2") << string_t("2")), "cond = `2 items for 2 dollars each'");

    TS_ASSERT_EQUALS(string_t(format(str) << string_t("0000000000000") << string_t("0000000000000")), "cond = `0000000000000 items for 0000000000000 dollars each'");
    TS_ASSERT_EQUALS(string_t(format(str) << string_t("0000000000000") << string_t("0000000000001")), "cond = `0000000000000 items for 0000000000001 dollar each'");
    TS_ASSERT_EQUALS(string_t(format(str) << string_t("0000000000000") << string_t("0000000000002")), "cond = `0000000000000 items for 0000000000002 dollars each'");
    TS_ASSERT_EQUALS(string_t(format(str) << string_t("000000000001x") << string_t("0004294967297")), "cond = `000000000001x item for 0004294967297 dollars'");

    string_t one = "11";
    TS_ASSERT_EQUALS(string_t(format(str) << one << one), "cond = `11 items for 11 dollars each'");
    one.erase(1);
    TS_ASSERT_EQUALS(string_t(format(str) << one << one), "cond = `1 item for 1 dollar'");
}
